extern crate pancurses;
use self::pancurses::{
    initscr, endwin, noecho, has_colors, start_color, init_pair,
    Window,
    COLOR_GREEN, COLOR_BLACK
};

extern crate rand;
use self::rand::Rng;
use self::rand::prelude::SliceRandom;

use std::{thread, time};

use config::{Config};
use util::{copy_shuffle};

// Board control and interaction. The view & controller.

mod board {}

const DEFAULT_ROWS: usize = 17;
const DEFAULT_COLS: usize = 24;

// Different actions are printed to the console at different speeds (the terminal displaying output, vs the player "typing")
const TERMINAL_PRINTING_SPEED: time::Duration = time::Duration::from_millis(20);
const TYPING_SPEED: time::Duration = time::Duration::from_millis(70);

// Everything displayed to the player, + a reference to the curses window.
#[derive(Debug)]
pub struct Screen {
    pub w: Window,
    attempts_left: u8,
    b: Board
}

// The game board.
#[derive(Debug)]
struct Board {
    rows: usize,
    cols: usize,
    c: Vec<char>
}

// The position of a word on the game board.
#[derive(Debug)]
struct Coord {
    start: usize,
    end: usize
}

impl Coord {
    fn new(start: usize, len: usize) -> Coord {
        Coord{start:start, end:start+len-1}
    }
}

impl Board {
    fn new(rows: Option<usize>, cols: Option<usize>) -> Board {
        let r = match rows {
            Some(r) => r,
            None => DEFAULT_ROWS
        };
        let c = match cols {
            Some(c) => c,
            None => DEFAULT_COLS
        };
        Board {
            rows: r,
            cols: c,
            c: Board::garbage_vec(r, c)
        }
    }

    fn garbage_vec(rows: usize, cols: usize) -> Vec<char> {
        let g: Vec<char> = "!@#$%^*()_-=+\\|/[]{}?\"\':;,.<>".chars().collect();
        let mut b = Vec::with_capacity(rows * cols);
        let mut rng = rand::thread_rng();

        for _ in 0..rows * cols {
            b.push(*g.choose(&mut rng).unwrap());
        }

        b
    }

    fn capacity(&self) -> usize {
        self.c.len()
    }

    fn get(&self, i: usize) -> char {
        self.c[i]
    }

    fn getxy(&self, x: usize, y: usize) -> char {
        self.c[x*y]
    }

    fn set(&mut self, i: usize, c: char) {
        self.c[i] = c;
    }

    fn setxy(&mut self, x: usize, y: usize, c: char) {
        self.c[x*y] = c;
    }

    // Think of the board in word_len "chunks": place a slot for a word in each chunk
    fn place_words(&self, num_of_words: usize, word_len: usize) -> Vec<Coord> {
        let chunk_len = self.capacity() / num_of_words;
        if chunk_len < word_len+1 {
            panic!("Can't fill board");
        }
        let mut coords: Vec<Coord> = Vec::with_capacity(num_of_words);
        let mut starting_coord = 0;
        for _ in 0..num_of_words {
            let coord = Coord::new(starting_coord, word_len);
            coords.push(coord);
            starting_coord+=chunk_len;
        }
        coords
    }

    // "shake" each word slot around, in a random order, 3 times/word.
    fn shake_words(&self, word_coords: &mut Vec<Coord>) {
        let mut rng = rand::thread_rng();

        for _ in 0..3 {
            let mut positions: Vec<usize> = (0..word_coords.len()).collect();
            positions.shuffle(&mut rng);
            for i in positions {
                let coord = &word_coords[i];
                // Where can this coord move (leaving at least 1 space between words)?
                let mut movement_options = Vec::with_capacity(2);

                let mut start = coord.start as i32;
                let mut end = -2;
                if i > 0 {
                    let prev_coord = &word_coords[i-1];
                    end = prev_coord.end as i32;
                }
                if start - end > 1 {
                    // It can move backwards
                    movement_options.push((start-end-1) * -1);
                }

                start = self.capacity() as i32;
                end = coord.end as i32;
                if i < word_coords.len()-1 {
                    let next_coord = &word_coords[i+1];
                    start = next_coord.start as i32;
                }
                if start - end > 1 {
                    // It can move forwards
                    movement_options.push(start-end-1);
                }

                // Pick a movement option, & pick some amount to move
                let mv = match movement_options.choose(&mut rng) {
                    None => continue,
                    Some(m) => if *m > 0 {
                        rng.gen_range(0, *m)
                    } else {
                        rng.gen_range(*m, 1)
                    }
                };
                word_coords[i] = Coord{
                    start: (coord.start as i32+mv) as usize,
                    end: (coord.end as i32+mv) as usize,
                };
            }
        }
    }
}

impl Screen {
    pub fn new(conf: &Config) -> Screen {
        let window = initscr();
        noecho();
        window_enable_colors();
        Screen {
            w: window,
            attempts_left: 4,
            b: Board::new(conf.rows, conf.cols)
        }
    }

    pub fn intro(&self) {
        self.w.clear();
        thread::sleep(time::Duration::from_millis(250));
        self.print_with_delay(0, String::from("WELCOME TO ROBCO INDUSTRIES (TM) TERMLINK"), true, TERMINAL_PRINTING_SPEED);
        self.move_cursor_to(1, 0);
        thread::sleep(time::Duration::from_millis(30));

        // TODO the rest of the intro

    }

    pub fn initialize_game(&self, conf: Config) {
        self.print_with_delay(0, String::from("ROBCO INDUSTRIES (TM) TERMLINK PROTOCOL"), false, TERMINAL_PRINTING_SPEED);
        self.print_with_delay(1, String::from("ENTER PASSWORD NOW"), false, TERMINAL_PRINTING_SPEED);
        self.print_with_delay(
            3,
            self.attempts_display(),
            false,
            TERMINAL_PRINTING_SPEED
        );
        let border_hex = border_hex_gen(self.b.rows);

        let shuffled_words = copy_shuffle(&conf.words);
        let words = &shuffled_words[..conf.choose];

        // Assumes validate() ensured `words` is not empty, and each word is the same length
        let word_len = words[0].len();
        let required_tiles = word_len * words.len() + (words.len() - 1);
        let board_tiles = self.b.capacity();
        if required_tiles > board_tiles {
            panic!("{} words of length {} requires {} tiles, but an {}x{} board only allows for {} tiles", conf.words.len(), word_len, required_tiles, self.b.cols, self.b.rows, board_tiles);
        }

        let mut coords = self.b.place_words(words.len(), word_len);
        self.b.shake_words(&mut coords);

        println!("{:?}", coords);
        println!("{:?}", self.b);

    }

    pub fn end_window(&self) {
        endwin();
    }

    fn move_cursor_to(&self, y: i32, x: i32) {
        self.w.mv(y, x);
        self.w.refresh();
    }

    fn print_with_delay(&self, line: i32, string: String, skippable: bool, delay: time::Duration) {
        self.w.mv(line, 0);
        for c in string.chars() {
            self.w.addch(c);
            self.w.refresh();
            if skippable {
                // TODO keyboard skip
            }
            thread::sleep(delay);
        }
    }

    fn attempts_display(&self) -> String {
        format!("{} ATTEMPT(S) LEFT:{}", self.attempts_left, " *".repeat(self.attempts_left as usize))
    }

}

fn window_enable_colors() {
    if has_colors() {
        start_color();
        init_pair(1, COLOR_GREEN, COLOR_BLACK);
    }
}

fn border_hex_gen(rows: usize) -> Vec<u32> {
    // Build the hex values, printed alongside the text and garbage.
    // The vector will be of length rows * 2
    // TODO: this would make a nice generator, once those are stable.
    let mut rng = rand::thread_rng();
    let mut hex = rng.gen_range(0,200) + 63744;

    let mut ls = Vec::with_capacity(rows * 2);
    for _ in 0..rows * 2 {
        ls.push(hex);
        hex += 12;
    }

    ls
}
